#!/usr/bin/env python
# coding: utf-8

# # 1D Discontinuous Galerkin Shallow Water Solver
#
# We solve the 1D Shallow Water Equations in conservative form:
#
# \begin{align*}
#     h_t + q_x &= 0 \\
#     q_t + \left[ \frac{q^2}{h} + \frac{1}{2} g h^2 \right]_x &= -g h b_x - C_f \left(\frac{q}{h}\right)^2
# \end{align*}
#
# We neglect friction so that $C_f = 0$.
#

import numpy as np
import matplotlib.pyplot as plt


import hedges.bc as bc
import hedges.fluxes as fluxes
import hedges.quadrature as quadrature
import hedges.rk as rk
import hedges.swe_1d.pde as pde


# Physical parameters
#
g = 1.0  # gravity

# Domain
#
tspan = (0.0, 4.0)
xspan = (-1, 1)

# Bathymetry parameters
#
b_smoothness = 0.1
b_amplitude = 0.02
b_slope = 0.05
assert(b_smoothness > 0)

# Inflow parameters
#
inflow_amplitude = 0.05
inflow_smoothness = 1.0
inflow_peak_time = 2.0
assert(inflow_amplitude > 0)

# Initial waveheight
#
h0 = 0.2
assert(h0 > 0)


def swe_bathymetry(x):
    """
    Describes bathymetry with an upslope which is perturbed by a hyperbolic tangent function.
    """
    return b_slope * x + b_amplitude * np.arctan(x / b_smoothness)


def swe_bathymetry_derivative(x):
    """
    Derivative of swe_bathymetry
    """
    return b_slope + b_amplitude / (
            b_smoothness * (1 + np.square(x / b_smoothness))
    )


def q_bc(t):
    """
    Describes a Gaussian inflow, where the function transitions to a constant value upon attaining
    its maximum value.

    :param t: Time
    :return:
    """
    t_np = np.array(t)
    return inflow_amplitude * np.exp(
        -np.square(
            np.minimum(t_np, inflow_peak_time * np.ones(t_np.shape)) - inflow_peak_time
        ) / (2 * np.square(inflow_smoothness))
    )


def initial_condition(x):
    """
    Creates initial conditions for (h, uh).

    :param x: Computational domain
    :return:
    """
    initial_height = h0 * np.ones(x.shape) - swe_bathymetry(x)  # horizontal water surface
    initial_flow = q_bc(0) * np.ones(x.shape)  # Start with inflow BC

    initial_values = np.array((
        initial_height,
        initial_flow,
    ))

    # Verify consistency of initial condition
    #
    if not np.allclose(initial_values[1][0], q_bc(0)):
        raise ValueError('Initial flow condition must match prescribed inflow.')

    return initial_values


# Plot bathymetry and ICs
#
xl, xr = xspan
t0, tf = tspan

xx = np.linspace(xl, xr, num=100)
tt = np.linspace(t0, tf, num=100)

fig, (h_ax, hv_ax, q_bc_ax) = plt.subplots(3, 1)

ic = initial_condition(xx)
bb = swe_bathymetry(xx)
qq_bc = q_bc(tt)

# Plot initial wave height and bathymetry
#
h_ax.plot(xx, ic[0] + bb)
h_ax.plot(xx, bb)
h_ax.set_title('Initial wave height $h(x, 0)$')

# Plot initial flow rate
#
hv_ax.plot(xx, ic[1])
hv_ax.set_title('Initial flow rate $q(x, 0)$')

# Plot flow rate at left boundary over simulation time
#
q_bc_ax.plot(tt, qq_bc)
q_bc_ax.set_title('Boundary flow rate $q({}, t)$'.format(xl))

plt.tight_layout()
plt.show()

# Instantiate solver with bathymetry
#
solver = pde.ShallowWater1D(
    b=swe_bathymetry,
    b_x=swe_bathymetry_derivative,
    gravity=g
)


t_interval_ms = 20
dt = t_interval_ms / 1000
surface_flux = fluxes.lax_friedrichs_flux
print('Integrating ODE system...')
solution = solver.solve(
    tspan=tspan,
    xspan=xspan,
    cell_count=16,
    polydeg=4,
    initial_condition=initial_condition,
    intercell_flux=surface_flux,
    left_boundary_flux=pde.ShallowWater1D.bc_prescribed_inflow(
        q_bc,
        gravity=g,
        surface_flux=surface_flux,
    ),
    right_boundary_flux=bc.transmissive_boundary(
        surface_flux=surface_flux,
        direction=bc.Direction.DOWNSTREAM,
    ),
    quad_rule=quadrature.gll,
    **{
        'method': rk.SSPRK33,
        't_eval': np.arange(tspan[0], tspan[1], dt),
        'max_step': dt,  # max time step for ODE solver
        'rtol': 1.0e-6,
        'atol': 1.0e-6,
    }
)

# Plot solution animation
#
ani, plt = solver.plot_animation(solution, frame_interval=t_interval_ms)

# Save animation to file
#
movie_name = 'swe_1d.gif'
print('Writing movie to {}...'.format(movie_name))

ani.save(movie_name, progress_callback=lambda i, n: print(
    f'Saving animation frame {i + 1}/{n}'
) if i % 50 == 0 else None)
print('Animation written to {}.'.format(movie_name))

plt.show()



import numpy as np
import scipy.optimize


def subcritical_height(gravity, b, q_l, h_r, b_r=0):
    """
    Computes the steady state height at set of points in the computational domain of
    subcritical Shallow Water flow over a bump.

    See Section 3.1.3 of https://doi.org/10.1002/fld.3741 for more information.

    :param gravity: Gravitational constant
    :param b: Array of bathymetry values at computational domain points
    :param q_l: Prescribed steady state flow at x = x_l (left boundary)
    :param h_r: Prescribed height at x = x_r (right boundary)
    :param b_r: Bathymetry height at x = x_r (right boundary)
    :return:
    """
    c = (b - (b_r + h_r) - np.square(q_l / h_r) / (2 * gravity))
    k = np.square(q_l) / (2 * gravity)

    return np.array([scipy.optimize.newton(
        lambda h: np.power(h, 3) + c[j] * np.square(h) + k,
        x0=h_r,
        maxiter=500
    ) for j, _ in enumerate(b)])



import enum

import numpy as np


from . import fluxes


class Direction(enum.Enum):
    UPSTREAM = 1
    DOWNSTREAM = 2


def transmissive_boundary(surface_flux=fluxes.lax_friedrichs_flux, direction=Direction.UPSTREAM):
    """
    Create a transmissive boundary by setting left and right ("ghost") states the same
    and computing the corresponding surface flux.

    For details see Section 9.2.5 of the book:
    - Eleuterio F. Toro (2001)
      Shock-Capturing Methods for Free-Surface Shallow Flows
      1st edition
      ISBN 0471987662

    :param surface_flux: Function used to compute numerical flux between adjacent cell states.
    :param direction: Boundary at which condition is implemented. Default is left boundary.
    """
    def _flux_function(u_l, u_r, xx, t, f, f_prime):
        if direction == Direction.UPSTREAM:
            return surface_flux(u_r, u_r, xx, t, f, f_prime)
        else:
            return surface_flux(u_l, u_l, xx, t, f, f_prime)

    return _flux_function


def reflective_boundary(surface_flux=fluxes.lax_friedrichs_flux, direction=Direction.UPSTREAM):
    """
    Create a transmissive boundary by reflecting the velocity between left and right ("ghost")
   and computing the corresponding surface flux.

    For details see Section 9.2.5 of the book:
    - Eleuterio F. Toro (2001)
      Shock-Capturing Methods for Free-Surface Shallow Flows
      1st edition
      ISBN 0471987662

    :param surface_flux: Function used to compute numerical flux between adjacent cell states.
    :param direction: Boundary at which condition is implemented. Default is left boundary.
    """
    def _flux_function(u_l, u_r, xx, t, f, f_prime):
        if direction == Direction.UPSTREAM:
            h_r, q_r = u_r
            return surface_flux(np.array((h_r, -q_r)), u_r, xx, t, f, f_prime)
        else:
            h_l, q_l = u_l
            return surface_flux(u_l, np.array((h_l, -q_l)), xx, t, f, f_prime)

    return _flux_function


def dirichlet_boundary(g, surface_flux=fluxes.lax_friedrichs_flux, direction=Direction.UPSTREAM):
    """
    Maintains a (possibly time-dependent) prescribed solution at the inflow boundary.

    :param g: A function with time as its first parameter, which should return the prescribed
              solution values at a particular time. The left and right cell interface values are
              passed to this function as named parameters.
    :param surface_flux: Function used to compute numerical flux between adjacent cell states.
    :param direction: Boundary at which condition is implemented. Default is left boundary.
    :return:
    """
    def _flux_function(u_l, u_r, xx, t, f, f_prime):
        # Evaluate prescribed solution at t
        #
        u = g(t, u_l=u_l, u_r=u_r)

        # Return numerical flux values
        #
        if direction == Direction.UPSTREAM:
            return surface_flux(np.array(u), u_r, xx, t, f, f_prime)
        else:
            return surface_flux(u_l, np.array(u), xx, t, f, f_prime)

    return _flux_function


def periodic_boundary(surface_flux=fluxes.lax_friedrichs_flux):
    """
    Periodic boundary conditions.

    :param surface_flux: Function used to compute numerical flux between adjacent cell states.
    :return:
    """
    def _flux_function(u_l, u_r, xx, t, f, f_prime):
        return surface_flux(u_l, u_r, xx, t, f, f_prime)

    return _flux_function



import numpy as np

from scipy.interpolate import lagrange


def _generate_lagrange_basis(x):
    """
    Generate a Lagrange basis consisting of characteristic polynomials of degree (len(x) - 1) for
    each collocation point. The j-th basis polynomial is computed by determining the interpolating
    polynomial for the points (x_i, delta_ij) where i represents the index of the collocation nodes
    and delta_ij is the Kronecker Delta function.

    :param x:
    :return:
    """
    basis = []
    for i in range(len(x)):
        samples = np.zeros(len(x))
        samples[i] = 1  # set to 1 to obtain a characteristic polynomial for the Lagrange basis
        basis.append(lagrange(x, samples))

    return basis


def lagrange_differentiation_matrix(x):
    """
    Computes the differentiation matrix for a Lagrange interpolating polynomial.

    :param x:
    :return:
    """
    basis = _generate_lagrange_basis(x)
    d = np.empty((len(x), len(x)))

    for j, b, in enumerate(basis):
        # b represents the j-th Lagrange basis function. We set the j-th column of the
        # differentiation matrix d to the derivative of this function evaluated at the provided
        # nodes.
        #
        d[:, j] = b.deriv()(x)

    return d


def gl(n, *args):
    """
    TODO: Figure out what changes need to be made to the hyperbolic solver for these points to work.
          You will be disappointed if you try to use these points in their current state (the points
          and weights are accurate).

    Computes N Gauss-Legendre nodes (intervals) on the interval [-1, 1] by using Numpy's built-in
    libraries for finding roots of Legendre polynomials.
    """
    # Get n-th Legendre polynomial basis function
    #
    legendre_basis_n = np.polynomial.Legendre.basis(n)

    # Find zeros of derivative of Legendre polynomials
    #
    x = legendre_basis_n.roots()

    # Compute weights
    #
    w = 2.0 / ((1 - np.square(x)) * (legendre_basis_n.deriv()(x))**2)

    return x, w


def gll(n, *args):
    """
    Computes N Gauss-Legendre-Lobatto nodes (intervals) on the interval [-1, 1] by using Numpy's
    built-in libraries for finding roots of differentiated Legendre polynomials.
    """
    # Get n-th Legendre polynomial basis function
    #
    legendre_basis_n = np.polynomial.Legendre.basis(n)

    # Find zeros of derivative of Legendre polynomials
    #
    x = legendre_basis_n.deriv().roots()
    x = np.insert(x, 0, -1.0)
    x = np.append(x, 1.0)

    # Compute weights
    #
    w = 2.0 / ((n*(n + 1)) * (legendre_basis_n(x)**2))

    return x, w


def _gll_newton_raphson(n, tol=10e-8):
    """
    Computes N Gauss-Legendre-Lobatto nodes (intervals) on the interval [-1, 1] by using the
    Newton-Raphson method to compute zeros of (1 - x^2)P_N'(x), where P_N' denotes the first
    derivative of the N-th degree Legendre polynomial.

    To use the Newton-Raphson method, we need d/dt [(1 - x^2)P_N'(x)].

    This function takes 1-2 orders of magnitude longer than the Numpy-based gll computation, but
    this function is included for reference as a secondary way to compute these points.

    See https://doi.org/10.1007/978-88-470-5522-3_10
    """

    # Initial guess
    #
    x = np.polynomial.chebyshev.chebpts2(n+1)

    # Legendre polynomials
    #
    legendre_bases = np.array((
        np.polynomial.Legendre.basis(n),
        np.polynomial.Legendre.basis(n-1),
    ))

    # Generate Legendre polynomials
    #
    x_prev = x
    for i in range(100):
        x = x + (legendre_bases[-1](x) - x * legendre_bases[0](x)) / ((n + 1) * legendre_bases[0](x))

        if np.linalg.norm(x - x_prev, np.inf) < tol:
            break

    # Compute weights
    #
    w = 2.0 / ((n*(n + 1)) * (legendre_bases[0](x)**2))

    return x, w


if __name__ == '__main__':
    """
    We add some basic testing here in order to verify agreement between the two GLL methods. This
    also provides a nice sanity check to compute GLL nodes/weights.
    
    """
    import time
    np.set_printoptions(suppress=True)
    print('Gauss-Legendre-Lobatto nodes/weights for 3rd-degree Legendre polynomial:')

    nodes, weights = gll(3)
    differentiation_matrix = lagrange_differentiation_matrix(nodes)

    print('x = {}'.format(nodes))
    print('w = {}'.format(weights))
    print('D = {}'.format(differentiation_matrix))

    # Compute GLL via Newton-Raphson
    #
    test_node_count = 100
    start = time.time()
    x_nr, w_nr = _gll_newton_raphson(test_node_count, 10e-8)
    end = time.time()
    print('Newton-Raphson took {} seconds for {} nodes'.format(end - start, test_node_count))
    start = time.time()

    # Compute GLL via Numpy
    #
    x_np, w_np = gll(test_node_count)
    end = time.time()
    print('Numpy differentiation took {} seconds for {} nodes'.format(end - start, test_node_count))

    # Assert equivalence
    #
    assert(np.all(np.isclose(x_nr, x_np)))
    assert(np.all(np.isclose(w_nr, w_np)))



import abc
import functools


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from scipy import integrate

from . import quadrature


class Hyperbolic1DSolver(abc.ABC):
    """
    This base class serves as a basic Discontinuous Galerkin solver for 1D hyperbolic PDEs and is
    written to be as clear as possible to those who are new to such solvers. To add support for a
    particular PDE (e.g., advection, SWE, etc.), the three abstract methods
    :func:`dg.hyperbolic_solver_1d.Hyperbolic1DSolver.hyperbolic_flux_term`,
    :func:`dg.hyperbolic_solver_1d.Hyperbolic1DSolver.hyperbolic_flux_jacobian`, and
    :func:`dg.hyperbolic_solver_1d.Hyperbolic1DSolver.source_term` defined in this class must be
    implemented. See :class:`dg.swe_1d.ShallowWater1D` for an example.

    This class also contains a basic function for animating numerical solutions; by default, all
    solution variables are plotted as subplots in a single animation, though this behavior can be
    overridden as desired by the subclasses implementation. See :class:`dg.swe_1d.ShallowWater1D`
    for an example.
    """
    solution_vars = NotImplemented  # Indicates the number of variables in the PDE solution.

    @abc.abstractmethod
    def hyperbolic_flux_term(self, u, xx, t):
        """
        Compute flux term from solution values (u) at a particular time step. This corresponds to
        F(u) in the conservative PDE form:

        u_t + [F(u)]_x = S(u)

        :param u: Solution values at t
        :param xx: Domain values
        :param t: Current timestep value
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def hyperbolic_flux_jacobian(self, u, xx, t):
        """
        Compute flux Jacobian from solution values (u) at a particular time step. This corresponds
        to F'(u) in the quasi-linear form:

        u_t + F'(u) u_x = S(u)

        :param u: Solution values at t
        :param xx: Domain values
        :param t: Current timestep value
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def source_term(self, u, xx, t):
        """
        Compute source term from solution values (u) at a particular time step. This corresponds to
        S(u) in the conservative PDE form:

        u_t + [F(u)]_x = S(u)

        :param u: Solution values at t
        :param xx: Domain values
        :param t: Current timestep value
        :return:
        """
        raise NotImplementedError

    @staticmethod
    def _to_physical_domain(x_l, x_r, xi):
        """
        Transforms a set of reference coordinates (i.e., points between [-1, 1]) from a cell's
        reference domain to a particular element in the physical domain.

        In other words, for a cell Q_l with bounds x_l and x_r, this function
        makes the affine transformation:

        [-1, 1] -> [x_l, x_r].

        :param x_l: Left boundary of physical domain cell
        :param x_l: Right boundary of physical domain cell
        :param xi: Collection of points contained in the reference domain [-1, 1].
        :return: Tuple of (x_k, J_k), where x_k are the mapped points and J_k is the Jacobian of
                 the transformation variables (used for quadrature evaluations on arbitrary
                 intervals)
        """
        return (x_l + x_r + xi * (x_r - x_l))/2, (x_r - x_l)/2

    def solve(self, tspan, xspan, cell_count, polydeg, left_boundary_flux, right_boundary_flux,
              intercell_flux, initial_condition, t_eval_interval=20,
              quad_rule=quadrature.gll, **ivp_options):
        """
        Solves the hyperbolic PDE via the Method of Lines (MoL) approach with standard SciPy ODE
        solvers.

        :param tspan: Time domain on which the PDE is to be solved.
        :param xspan: Physical domain on which the PDE is to be solved.
        :param cell_count: Number of cells to split domain into
        :param polydeg: Degree of approximating polynomial (same across all cells).
        :param left_boundary_flux: Function used to compute flux at left boundary (inflow)
        :param right_boundary_flux: Function used to compute flux at right boundary (outflow)
        :param intercell_flux: Function used to compute flux in between adjacent cells.
        :param initial_condition: Function used to compute the solution's initial condition.
        :param method: ODE Integration scheme used to solve semidiscretized problem. Should
                       correspond to an existing SciPy ODE solver.
        :param t_eval_interval: Interval (in ms) to evaluate interpolated solution. For instance,
                                if tspan=[0, 0.5] and this value is set to 100, this function
                                returns the solution at points {0, 0.1, 0.2, 0.3, 0.4, 0.5}.
        :param quad_rule: Quadrature rule. Given a polynomial degree, this function should
                          return a tuple (x, w) where x and w are arrays containing the
                          quadrature nodes and weights.
        :return: (sol, xx) tuple, where sol is the system solution and xx is the domain (in the
                 shape that each solution variable at each timestep should be reshaped to). The
                 first dimension represents a particular cell and the second the GLL nodes (as
                 mapped from the reference domain [-1, 1] into the physical domain) in each cell.
        """
        x_l, x_r = xspan
        t_0, t_f = tspan

        # Gimme some SWEet Gauss-Legendre-Lobatto points/weights on [-1, 1]
        #
        x, w = quad_rule(polydeg)

        # Get Lagrange differentiation matrix to compute derivative of polynomial interpolant
        #
        differentiation_matrix = quadrature.lagrange_differentiation_matrix(x)

        # Mass matrix; basically just a convenient way to implement a Gaussian quadrature rule.
        #
        mass_matrix = np.diag(w)

        # Boundary matrix for assigning numerical flux values
        #
        boundary_matrix = np.zeros(mass_matrix.shape)
        boundary_matrix[0, 0] = -1
        boundary_matrix[-1, -1] = 1

        # First dimension is the solution variable of interest (e.g., length 1 for scalar 1D
        # advection, length 2 for 1D SWE, etc.), the second dimension represents the domain
        # cells, and the third dimension represents the cell (GLL) nodes.
        #
        u_shape = (self.solution_vars, cell_count, len(x))

        # Shape of domain is a tuple of the latter two elements of u_shape.
        #
        xx_shape = u_shape[1:]

        # Split domain into evenly-spaced cells
        #
        dx = (x_r - x_l) / cell_count

        # Stores node values across entire computational domain (i.e., [x_k^L, x_k^R])
        #
        xx = np.empty(xx_shape)

        # Stores the length of each domain cell.
        #
        cell_lengths = np.empty(cell_count)

        for cell_k in range(0, cell_count):
            cell_left = x_l + cell_k * dx
            cell_right = cell_left + dx
            xx[cell_k], cell_lengths[cell_k] = self._to_physical_domain(cell_left, cell_right, x)

        # This allows us to get a feel for the status of the integration scheme. This will print our
        # the status no more than 20 times over the course of integration.
        #
        time_buckets = np.linspace(t_0, t_f, num=20)

        def uprime(t, u_col, ivp_args):
            if t > time_buckets[ivp_args['bucket']]:
                ivp_args['bucket'] = ivp_args['bucket'] + 1
                print('Computing du/dt at t={}'.format(t))

            # Reshape column vector into cell-node matrix.
            #
            u = u_col.reshape(u_shape)

            # Stores computed value of u' at this time step
            #
            du_t = np.zeros(u_shape)

            # Placeholder for numerical flux values
            #
            numerical_flux = np.zeros(u_shape)

            # Evaluate hyperbolic flux
            #
            f_u = self.hyperbolic_flux_term(u, xx, t)

            # Evaluate hyperbolic source term
            #
            s_u = self.source_term(u, xx, t)

            for cell_k in range(1, cell_count - 1):
                left_flux = intercell_flux(
                    u[:, cell_k - 1, -1],
                    u[:, cell_k, 0],
                    xx,
                    t,
                    f=self.hyperbolic_flux_term,
                    f_prime=self.hyperbolic_flux_jacobian
                )

                numerical_flux[:, cell_k, 0] = left_flux
                numerical_flux[:, cell_k - 1, -1] = left_flux

                right_flux = intercell_flux(
                    u[:, cell_k, -1],
                    u[:, cell_k + 1, 0],
                    xx,
                    t,
                    f=self.hyperbolic_flux_term,
                    f_prime=self.hyperbolic_flux_jacobian
                )
                numerical_flux[:, cell_k, -1] = right_flux
                numerical_flux[:, cell_k + 1, 0] = right_flux

            # Compute and assign numerical flux for boundary interfaces. This is where boundary
            # conditions are implemented.
            #
            numerical_flux[:, 0, 0] = left_boundary_flux(
                u[:, -1, -1],
                u[:, 0, 0],
                xx,
                t,
                f=self.hyperbolic_flux_term,
                f_prime=self.hyperbolic_flux_jacobian
            )
            numerical_flux[:, -1, -1] = right_boundary_flux(
                u[:, -1, -1],
                u[:, 0, 0],
                xx,
                t,
                f=self.hyperbolic_flux_term,
                f_prime=self.hyperbolic_flux_jacobian
            )

            # Compute integrals via quadrature
            #
            for var in range(0, self.solution_vars):
                # Update values in each cell
                #
                for cell_k in range(0, cell_count):
                    # Compute volume integrals
                    #
                    du_t[var][cell_k] += np.linalg.solve(mass_matrix, differentiation_matrix.T) \
                                         @ mass_matrix \
                                         @ f_u[var][cell_k]

                    # Compute surface integrals
                    #
                    du_t[var][cell_k] -= np.linalg.solve(mass_matrix, boundary_matrix) \
                                         @ (numerical_flux[var][cell_k])

                    # Apply Jacobian from mapping to reference element
                    #
                    du_t[var][cell_k] /= cell_lengths[cell_k]

                    # Add source term
                    #
                    du_t[var][cell_k] += s_u[var][cell_k]

            # Flatten column-wise
            #
            return du_t.flatten()

        # Flatten initial condition cell grid into vector
        #
        ic = initial_condition(xx).flatten()

        # We wrap this value in a dict so that uprime can actually mutate t_uniq. t_uniq is used
        # to print ODE solver progress.
        #
        ode_args = {'bucket': 0}

        # Solve ODE system
        #
        sol = integrate.solve_ivp(
            uprime,
            (t_0, t_f),
            ic,
            args=(ode_args,),
            **ivp_options
        )

        return sol, xx

    def plot_animation(self, solution, frame_interval=20, animate_frame=None, initialize_plots=None):
        """
        Animates plots of time-dependent solutions over time. This function plots all solution
        variables as subplots in a single plot.

        :param solution: Solution tuple in the form returned by
                         :func:`~hyperbolic_solver_1d.Hyperbolic1DSolver.solve`
        :param frame_interval: Interval between animation frames in milliseconds
        :param animate_frame: Optional custom animation function (e.g., to draw SWE bathymetry)
        :param initialize_plots: Optional function for initializing plots. If not provided,
                                 solutions will be plotted according to their respective min/max
                                 values over the entire time domain.
        :return:
        """
        sol, xx = solution

        # Plot animated solution
        #
        u_shape = (self.solution_vars, xx.shape[0], xx.shape[1])
        fig, ax = plt.subplots(self.solution_vars, 1)
        fig.tight_layout()

        if initialize_plots:
            lines = initialize_plots(ax, xx, sol, u_shape)
        else:
            lines = self._initialize_plots(ax, xx, sol, u_shape)

        ani = animation.FuncAnimation(
            fig,
            functools.partial(
                animate_frame or self._default_frame_animation,
                domain=xx,
                soln=sol,
                plot_lines=lines,
                axes=ax,
                soln_shape=u_shape,
            ),
            frames=len(sol.t),
            interval=frame_interval,
            repeat=True,
            repeat_delay=1.0 * 1000,
        )

        return ani, plt

    def _initialize_plots(self, axes, domain, soln, soln_shape):
        """
        Initializes plots for animation. Axis limits for each subplot are set to max/min values
        across entire time domain with respect to the subplot's solution variable.

        :param axes: Subplot axes.
        :param domain:
        :param soln: Solution indexed by time step
        :param soln_shape: Shape of solution variable
        :return:
        """
        plot_lines = []

        # Reshape solution into something a little more intuitive to work with. First dimension
        # is the time step index, second is the solution variable of interest, third dimension
        # is the cell index, and fourth is the node.
        #
        uu = soln.y.T.reshape(len(soln.t), *soln_shape)

        for i in range(0, self.solution_vars):
            axes[i].set_ylim(uu[:, i].min(), uu[:, i].max())
            line, = axes[i].plot(domain.flatten(), uu[0, i].flatten())
            plot_lines.append(line)

        return plot_lines

    def _default_frame_animation(self, n, axes, plot_lines, domain, soln, soln_shape):
        """
        Animate plot according to:

        https://matplotlib.org/stable/api/_as_gen/matplotlib.animation.FuncAnimation.html

        :param n: Frame index
        :param axes: Subplot axes.
        :param plot_lines: Plot lines to update. This should contain one line per solution.
        :param domain:
        :param soln: Solution indexed by time step
        :param soln_shape: Shape of solution variable
        :return:
        """
        u_i = soln.y[:, n].reshape(soln_shape)

        # Plot each solution quantity
        #
        for k, line in enumerate(plot_lines):
            axes[k].set_title('$q_{}; t = {:0.2f}$'.format(k+1, soln.t[n]))
            line.set_xdata(domain.flatten())
            line.set_ydata(u_i[k].flatten())

        # This is the format required by the animate callback for plots.
        #
        return [(line,) for line in plot_lines]
